@@ -1,10 +1,10 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1996-2003
+ * Copyright (c) 1996-2004
  *	Sleepycat Software.  All rights reserved.
  *
- * $Id: mutex.h,v 11.90 2003/09/20 21:40:49 bostic Exp $
+ * $Id: mutex.h,v 11.100 2004/10/05 14:41:12 mjc Exp $
  */
 
 #ifndef _DB_MUTEX_H_
@@ -128,6 +128,11 @@ extern void _spin_unlock(tsl_t *);
 
 #ifndef	MUTEX_ALIGN
 #define	MUTEX_ALIGN	16
+#ifdef __hpux__
+# define	HPUX_MUTEX_PAD	 8
+#else
+# define	HPUX_MUTEX_PAD	 4
+#endif
 #endif
 #endif
 
@@ -290,8 +295,7 @@ typedef SEM_ID tsl_t;
  * trying to initialize the global lock at the same time.
  */
 #undef	DB_BEGIN_SINGLE_THREAD
-#define	DB_BEGIN_SINGLE_THREAD						\
-do {									\
+#define	DB_BEGIN_SINGLE_THREAD do {					\
 	if (DB_GLOBAL(db_global_init))					\
 		(void)semTake(DB_GLOBAL(db_global_lock), WAIT_FOREVER);	\
 	else {								\
@@ -341,13 +345,13 @@ typedef unsigned int tsl_t;
  *********************************************************************/
 #if defined(HAVE_MUTEX_WIN32) || defined(HAVE_MUTEX_WIN32_GCC)
 #define	MUTEX_FIELDS							\
-	LONG tas;							\
+	LONG volatile tas;						\
 	LONG nwaiters;							\
 	u_int32_t id;	/* ID used for creating events */		\
 
 #if defined(LOAD_ACTUAL_MUTEX_CODE)
-#define	MUTEX_SET(tsl)		(!InterlockedExchange((PLONG)tsl, 1))
-#define	MUTEX_UNSET(tsl)	(*(tsl) = 0)
+#define	MUTEX_SET(tsl)		(InterlockedExchange((PLONG)(tsl), 1) == 0)
+#define	MUTEX_UNSET(tsl)	(InterlockedExchange((PLONG)(tsl), 0))
 #define	MUTEX_INIT(tsl)		MUTEX_UNSET(tsl)
 
 /*
@@ -360,8 +364,10 @@ typedef unsigned int tsl_t;
  * platforms, and it improves performance on Pentium 4 processor platforms."
  */
 #ifdef HAVE_MUTEX_WIN32
+#ifndef _WIN64
 #define	MUTEX_PAUSE		{__asm{_emit 0xf3}; __asm{_emit 0x90}}
 #endif
+#endif
 #ifdef HAVE_MUTEX_WIN32_GCC
 #define	MUTEX_PAUSE		asm volatile ("rep; nop" : : );
 #endif
@@ -375,9 +381,8 @@ typedef unsigned int tsl_t;
 typedef unsigned char tsl_t;
 
 #ifdef LOAD_ACTUAL_MUTEX_CODE
-/*
- * For gcc/68K, 0 is clear, 1 is set.
- */
+#define	MUTEX_SET_TEST	1		/* gcc/68K: 0 is clear, 1 is set. */
+
 #define	MUTEX_SET(tsl) ({						\
 	register tsl_t *__l = (tsl);					\
 	int __r;							\
@@ -467,9 +472,8 @@ typedef volatile u_int32_t tsl_t;
 typedef unsigned char tsl_t;
 
 #ifdef LOAD_ACTUAL_MUTEX_CODE
-/*
- * For arm/gcc, 0 is clear, 1 is set.
- */
+#define	MUTEX_SET_TEST	1		/* gcc/arm: 0 is clear, 1 is set. */
+
 #define	MUTEX_SET(tsl) ({						\
 	int __r;							\
 	asm volatile(							\
@@ -486,46 +490,6 @@ typedef unsigned char tsl_t;
 #endif
 #endif
 
-/*********************************************************************
- * MIPS/gcc assembly.
- *********************************************************************/
-#ifdef HAVE_MUTEX_MIPS_GCC_ASSEMBLY
-typedef u_int32_t tsl_t;
-
-#ifndef        MUTEX_ALIGN
-#define        MUTEX_ALIGN     4
-#endif
-
-#ifdef LOAD_ACTUAL_MUTEX_CODE
-/*
- * For gcc/mips.  Should return 0 if could not acquire the lock, 1 if
- * lock was acquired properly.
- */
-static inline int
-MUTEX_SET(tsl_t *tsl) {
-       register tsl_t *__l = tsl;
-       register tsl_t __r;
-       __asm__ __volatile__(
-               "       .set push       \n"
-               "       .set mips2      \n"
-               "       .set noreorder  \n"
-               "       .set nomacro    \n"
-               "1:     ll      %0, %1          \n"
-               "       bne     %0, $0, 1f      \n"
-               "        xori   %0, %0, 1       \n"
-               "       sc      %0, %1          \n"
-               "       beql    %0, $0, 1b      \n"
-               "        xori   %0, 1           \n"
-               "1:     .set pop                  "
-               : "=&r" (__r), "+R" (*__l));
-       return __r;
-}
-
-#define        MUTEX_UNSET(tsl)        (*(volatile tsl_t *)(tsl) = 0)
-#define        MUTEX_INIT(tsl)         MUTEX_UNSET(tsl)
-#endif
-#endif
-
 /*********************************************************************
  * HPPA/gcc assembly.
  *********************************************************************/
@@ -534,6 +498,11 @@ typedef u_int32_t tsl_t;
 
 #ifndef	MUTEX_ALIGN
 #define	MUTEX_ALIGN	16
+#ifdef __hpux__
+# define	HPUX_MUTEX_PAD	 8
+#else
+# define	HPUX_MUTEX_PAD	 4
+#endif
 #endif
 
 #ifdef LOAD_ACTUAL_MUTEX_CODE
@@ -549,7 +518,7 @@ typedef u_int32_t tsl_t;
 	__r & 1;							\
 })
 
-#define	MUTEX_UNSET(tsl)	(*(tsl) = -1)
+#define	MUTEX_UNSET(tsl)        (*(volatile tsl_t *)(tsl) = -1)
 #define	MUTEX_INIT(tsl)		(MUTEX_UNSET(tsl), 0)
 #endif
 #endif
@@ -561,9 +530,8 @@ typedef u_int32_t tsl_t;
 typedef unsigned char tsl_t;
 
 #ifdef LOAD_ACTUAL_MUTEX_CODE
-/*
- * For gcc/ia64, 0 is clear, 1 is set.
- */
+#define	MUTEX_SET_TEST	1		/* gcc/ia64: 0 is clear, 1 is set. */
+
 #define	MUTEX_SET(tsl) ({						\
 	register tsl_t *__l = (tsl);					\
 	long __r;							\
@@ -618,11 +586,12 @@ typedef u_int32_t tsl_t;
  * 'set' mutexes have the value 1, like on Intel; the returned value from
  * MUTEX_SET() is 1 if the mutex previously had its low bit clear, 0 otherwise.
  */
-#ifdef HAVE_MUTEX_PPC_GCC_ASSEMBLY
+#define	MUTEX_SET_TEST	1		/* gcc/ppc: 0 is clear, 1 is set. */
+
 static inline int
 MUTEX_SET(int *tsl)  {
          int __r;
-         asm volatile (
+    asm volatile (
 "0:                             \n\t"
 "       lwarx   %0,0,%1         \n\t"
 "       cmpwi   %0,0            \n\t"
@@ -639,7 +608,7 @@ MUTEX_SET(int *tsl)  {
          : "cr0", "memory");
          return (int)tsl;
 }
-#endif
+
 static inline int
 MUTEX_UNSET(tsl_t *tsl) {
          asm volatile("sync" : : : "memory");
@@ -656,7 +625,7 @@ MUTEX_UNSET(tsl_t *tsl) {
 typedef int tsl_t;
 
 #ifndef	MUTEX_ALIGN
-#define	MUTEX_ALIGN	 sizeof(int)
+#define	MUTEX_ALIGN	sizeof(int)
 #endif
 
 #ifdef LOAD_ACTUAL_MUTEX_CODE
@@ -677,9 +646,8 @@ typedef int tsl_t;
 typedef int tsl_t;
 
 #ifdef LOAD_ACTUAL_MUTEX_CODE
-/*
- * For gcc/S390, 0 is clear, 1 is set.
- */
+#define	MUTEX_SET_TEST	1		/* gcc/S390: 0 is clear, 1 is set. */
+
 static inline int
 MUTEX_SET(tsl_t *tsl) {							\
 	register tsl_t *__l = (tsl);					\
@@ -709,9 +677,8 @@ typedef unsigned char tsl_t;
 #ifdef LOAD_ACTUAL_MUTEX_CODE
 /*
  * UnixWare has threads in libthread, but OpenServer doesn't (yet).
- *
- * For cc/x86, 0 is clear, 1 is set.
  */
+#define	MUTEX_SET_TEST	1		/* cc/x86: 0 is clear, 1 is set. */
 
 #if defined(__USLC__)
 asm int
@@ -751,9 +718,9 @@ typedef unsigned char tsl_t;
  * so is functional there as well.  For v7, stbar may generate an illegal
  * instruction and we have no way to tell what we're running on.  Some
  * operating systems notice and skip this instruction in the fault handler.
- *
- * For gcc/sparc, 0 is clear, 1 is set.
  */
+#define	MUTEX_SET_TEST	1		/* gcc/sparc: 0 is clear, 1 is set. */
+
 #define	MUTEX_SET(tsl) ({						\
 	register tsl_t *__l = (tsl);					\
 	register tsl_t __r;						\
@@ -785,6 +752,46 @@ typedef int tsl_t;
 #endif
 #endif
 
+/*********************************************************************
+ * MIPS/gcc assembly.
+ *********************************************************************/
+#ifdef HAVE_MUTEX_MIPS_GCC_ASSEMBLY
+typedef u_int32_t tsl_t;
+
+#ifndef MUTEX_ALIGN
+#define MUTEX_ALIGN sizeof(u_int32_t)
+#endif
+
+#ifdef LOAD_ACTUAL_MUTEX_CODE
+/*
+ * For gcc/MIPS.  Should return 0 if could not acquire the lock, 1 if
+ * lock was acquired properly.
+ */
+static inline int
+MUTEX_SET(tsl_t *tsl) {
+       register tsl_t *__l = tsl;
+       register tsl_t __r;
+       __asm__ __volatile__(
+               "       .set push           \n"
+               "       .set mips2          \n"
+               "       .set noreorder      \n"
+               "       .set nomacro        \n"
+               "1:     ll      %0, %1      \n"
+               "       bne     %0, $0, 1f  \n"
+               "        xori   %0, %0, 1   \n"
+               "       sc      %0, %1      \n"
+               "       beql    %0, $0, 1b  \n"
+               "        xori   %0, 1       \n"
+               "1:     .set pop              "
+               : "=&r" (__r), "+R" (*__l));
+       return __r;
+}
+
+#define	MUTEX_UNSET(tsl)        (*(volatile tsl_t *)(tsl) = 0)
+#define	MUTEX_INIT(tsl)         MUTEX_UNSET(tsl)
+#endif
+#endif
+
 /*********************************************************************
  * x86/gcc assembly.
  *********************************************************************/
@@ -792,9 +799,8 @@ typedef int tsl_t;
 typedef unsigned char tsl_t;
 
 #ifdef LOAD_ACTUAL_MUTEX_CODE
-/*
- * For gcc/x86, 0 is clear, 1 is set.
- */
+#define	MUTEX_SET_TEST	1		/* gcc/x86: 0 is clear, 1 is set. */
+
 #define	MUTEX_SET(tsl) ({						\
 	register tsl_t *__l = (tsl);					\
 	int __r;							\
@@ -805,7 +811,7 @@ typedef unsigned char tsl_t;
 	__r & 1;							\
 })
 
-#define	MUTEX_UNSET(tsl)	(*(tsl) = 0)
+#define	MUTEX_UNSET(tsl)        (*(volatile tsl_t *)(tsl) = 0)
 #define	MUTEX_INIT(tsl)		MUTEX_UNSET(tsl)
 
 /*
@@ -845,7 +851,7 @@ typedef unsigned char tsl_t;
 
 /*
  * !!!
- * The flag arguments for __db_mutex_setup (and the underyling intialization
+ * The flag arguments for __db_mutex_setup (and the underlying initialization
  * function for the mutex type, for example, __db_tas_mutex_init), and flags
  * stored in the DB_MUTEX structure are combined, and may not overlap.  Flags
  * to __db_mutex_setup:
@@ -906,6 +912,11 @@ struct __mutex_t {
 	u_int32_t flags;		/* MUTEX_XXX */
 };
 
+/* Macro to clear mutex statistics. */
+#define	MUTEX_CLEAR(mp) {						\
+	(mp)->mutex_set_wait = (mp)->mutex_set_nowait = 0;		\
+}
+
 /* Redirect calls to the correct functions. */
 #ifdef HAVE_MUTEX_THREADS
 #if defined(HAVE_MUTEX_PTHREADS) ||					\
@@ -915,7 +926,8 @@ struct __mutex_t {
 #define	__db_mutex_lock(a, b)		__db_pthread_mutex_lock(a, b)
 #define	__db_mutex_unlock(a, b)		__db_pthread_mutex_unlock(a, b)
 #define	__db_mutex_destroy(a)		__db_pthread_mutex_destroy(a)
-#elif defined(HAVE_MUTEX_WIN32) || defined(HAVE_MUTEX_WIN32_GCC)
+#else
+#if defined(HAVE_MUTEX_WIN32) || defined(HAVE_MUTEX_WIN32_GCC)
 #define	__db_mutex_init_int(a, b, c, d)	__db_win32_mutex_init(a, b, d)
 #define	__db_mutex_lock(a, b)		__db_win32_mutex_lock(a, b)
 #define	__db_mutex_unlock(a, b)		__db_win32_mutex_unlock(a, b)
@@ -926,6 +938,7 @@ struct __mutex_t {
 #define	__db_mutex_unlock(a, b)		__db_tas_mutex_unlock(a, b)
 #define	__db_mutex_destroy(a)		__db_tas_mutex_destroy(a)
 #endif
+#endif
 #else
 #define	__db_mutex_init_int(a, b, c, d)	__db_fcntl_mutex_init(a, b, c)
 #define	__db_mutex_lock(a, b)		__db_fcntl_mutex_lock(a, b)
